"""
vula frontend constant values
"""
WIDTH = 1050
HEIGHT = 800

BACKGROUND_COLOR = '#333941'
BACKGROUND_COLOR_CARD = '#1E272F'
BACKGROUND_COLOR_ENTRY = '#515456'
BACKGROUND_COLOR_ERROR = '#890000'
TEXT_COLOR_HEADER = '#BB89EE'
TEXT_COLOR_HEADER_2 = '#C0BABA'
TEXT_COLOR_WHITE = '#FFFFFF'
TEXT_COLOR_BLACK = '#000716'
TEXT_COLOR_GREEN = '#5BC960'
TEXT_COLOR_RED = '#EA1C1C'
TEXT_COLOR_GREY = '#838383'
TEXT_COLOR_PURPLE = '#BB89EE'
TEXT_COLOR_YELLOW = '#FFB800'
TEXT_COLOR_ORANGE = '#EF884F'

FONT = 'Roboto Regular'
FONT_SIZE_HEADER = 36 * -1
FONT_SIZE_HEADER_2 = 24 * -1
FONT_SIZE_TEXT_XS = 11 * -1
FONT_SIZE_TEXT_S = 12 * -1
FONT_SIZE_TEXT_M = 13 * -1
FONT_SIZE_TEXT_L = 14 * -1
FONT_SIZE_TEXT_XL = 16 * -1
FONT_SIZE_TEXT_XXL = 20 * -1

IMAGE_BASE_PATH = '/usr/share/icons/vula/'
